#!/bin/sh

#
# iSCSI Target Configuration Updater
#
# by Steffen Plotner, 2005-2006
#

ROOT=$(dirname $0)

#
# online requires ietd to be running, offline just generates the config file
#
run_mode=$1

if [ -z "$run_mode" ]; then
	run_mode="online"
fi

if [ "$run_mode" != "offline" -a "$run_mode" != "online" ]; then
	echo "$0 [online|offline]"
	echo "Default runmode is online. Choose offline if IETD is not running"
	echo "and you still want to update the ietd.conf file"
	exit 1
fi

echo "MODE> $run_mode"

if [ $run_mode = "online" ]; then
	if [ ! -e /proc/net/iet/session -o ! -e /proc/net/iet/volume ]; then
		echo "ERR> iscsi target software is not loaded, execute as '$0 offline' instead"
		exit 1
	fi
fi

#
# overridable variables
#
LUN_TYPEIO=fileio

#
# include the user's common configuration and target definitions
#
. $ROOT/common.conf
. $ROOT/target.conf

#
# setup lunsize tracking directory
#
if [ ! -d $ROOT/lunsize ]; then
	mkdir $ROOT/lunsize
fi

#------------------------------------------------------------------------------
# get tid from target name
#
# stdout returns tid which can be
# 	0 if not found, 0 is not a legal tid
#	> 0
#
# return 0 on success, 1 on failure
#------------------------------------------------------------------------------
function get_tid_from_target
{
	local target_name
	target_name=$1
	
	if [ -z "$target_name" ]; then
		echo "0"
		return 1
	fi
	
	local line
	line=$(cat /proc/net/iet/session | grep "name:$target_name$")
	if [ -z "$line" ]; then
		echo "0"
	else
		# tid:1 name:test
		echo $line | cut -f1 -d' ' | cut -f2 -d:
	fi
	return 0
}

#------------------------------------------------------------------------------
# get next tid number
#
# stdout returns tid
#
# return 0 on success, 1 on failure
#------------------------------------------------------------------------------
function get_next_tid
{
	local last_tid
	last_tid=$(cat /proc/net/iet/session | \
		grep "^tid:" | \
		cut -f1 -d' ' | \
		cut -f2 -d: | \
		sort -n -r | \
		head -n 1)
	local new_tid
	new_tid=$(($last_tid + 1))
	echo "$new_tid"
	return 0
}

#------------------------------------------------------------------------------
# get tid's lun info
#------------------------------------------------------------------------------
function get_lun_info_from_tid
{
	local vf
	local tid
	local linecount
	local startline
	local lastN
	local remainder
	local nexttid_line
	
	vf=/proc/net/iet/volume
	tid=$1
	
	#
	# current line count
	#
	linecount=$(cat $vf | wc -l)
	if [ -z "$linecount" ]; then
		return
	fi
	
	#
	# starting line of tid:# (notice the space after the 'd'
	#
	startline=$(grep -n "^tid:$tid " $vf | head -n 1 | cut -d: -f1)
	if [ -z "$startline" ]; then
		return
	fi
	
	#
	# doing a tail, get the remaining lines based on startline
	#
	lastN=$(($linecount-$startline))
	remainder=$(tail -n $lastN $vf)
	
	#echo "$remainder"
	
	#
	# find next tid, get its line number, and get starting number
	# of lines-1
	#
	nexttid_line=$(echo "$remainder" | grep -n "^tid:" | head -n 1 | cut -d: -f1)
	if [ -z "$nexttid_line" ]; then
		if [ -n "$remainder" ]; then
			echo "$remainder"
		fi
	else
		#echo "nexttid=$nexttid_line"
		nexttid_line=$(($nexttid_line - 1))
		echo "$remainder" | head -n $nexttid_line
	fi
}

#------------------------------------------------------------------------------
# get lvm slice size (return NULL if undetermined)
#------------------------------------------------------------------------------
function get_lv_size
{
	lv_name=$1
	local cmd
	local lv_size
	local ccode
	
	cmd="lvs $lv_name -o lv_size --noheadings --units g"
	lv_size=$($cmd 2> /dev/null | awk '{print $1}')
	if [ -z "$lv_size" ]; then
		lv_size="NULL"
	fi
	echo "$lv_size"
	return 0
}

#------------------------------------------------------------------------------
# track lun size
#------------------------------------------------------------------------------
function track_lun_size_set
{
	lv_name=$1
	lv_size=$2
	
	file=$(echo $lv_name | sed -e 's/\//_/g')
	echo "$lv_size" > $ROOT/lunsize/$file
}

function track_lun_size_get
{
	lv_name=$1
	
	file=$(echo $lv_name | sed -e 's/\//_/g')
	if [ -e $ROOT/lunsize/$file ]; then
		cat $ROOT/lunsize/$file
	fi
}

#------------------------------------------------------------------------------
# enumerate the requested target's sessions and connections id and
# terminate them
#------------------------------------------------------------------------------
function delete_target_connections
{
	local FN
	local tid
	local state
	local ccode
	local cmd
	local sid
	local cid
	
	tid=$1
	state="search_tid"
	
	FN="iSCSI_DELETE_CONN"

	while read line; do

		# if we are done searching for tid, check that
		# if we encounter a different tid, we need to exit
		if [ $state != "search_tid" ]; then
			echo "$line" | grep "^tid:" > /dev/null
			ccode=$?
			if [ $ccode -eq 0 ]; then
				break
			fi
		fi

		case $state in
			search_tid)
				echo "$line" | grep "^tid:$tid " > /dev/null
				ccode=$?
				if [ $ccode -eq 0 ]; then
					echo "$FN> $line"
					state="search_sid"
				fi
				;;
			search_sid)
				echo "$line" | grep "^sid:" > /dev/null
				ccode=$?
				if [ $ccode -eq 0 ]; then
					echo "$FN> $line"
					sid=$(echo "$line" | cut -f1 -d' ' | cut -f2 -d:)
					state="search_cid"
				fi
				;;
			search_cid)
				echo "$line" | grep "^cid:" > /dev/null
				ccode=$?
				if [ $ccode -eq 0 ]; then
					echo "$FN> $line"
					cid=$(echo "$line" | cut -f1 -d' ' | cut -f2 -d:)
					
					cmd="ietadm --op delete --tid=$tid --sid=$sid --cid=$cid"
					echo "$FN> $cmd"
					$cmd
					
					state="search_sid"
				fi
				;;
		esac
		#echo "$line"
	done < /proc/net/iet/session
	
}

#------------------------------------------------------------------------------
# block/unblock target IO
#------------------------------------------------------------------------------
function target_io
{
	local ip_allow
	local on_off
	local old_IFS
	local FN
	local cmd
	
	FN="iSCSI_target_io"
	ip_allow=$1
	on_off=$2 
	
	old_IFS="$IFS"
	IFS=','
	for ip in $ip_allow; do
		IFS="$old_IFS"
		#
		# (un)block incoming traffic
		#
		cmd="iptables -D INPUT -p tcp --dport 3260 -s $ip -j DROP"
		if [ $on_off == "unblock" ]; then
			echo "$FN> $cmd"
		fi
		$cmd 2> /dev/null
		
		if [ $on_off == "block" ]; then
			cmd="iptables -I INPUT -p tcp --dport 3260 -s $ip -j DROP"
			echo "$FN> $cmd"
			$cmd
			sleep 2
		fi
		
		#
		# (un)block outgoing traffic
		#
		cmd="iptables -D OUTPUT -p tcp --sport 3260 -d $ip -j DROP"
		if [ $on_off == "unblock" ]; then
			echo "$FN> $cmd"
		fi
		$cmd 2> /dev/null

		if [ $on_off == "block" ]; then
			cmd="iptables -I OUTPUT -p tcp --sport 3260 -d $ip -j DROP"
			echo "$FN> $cmd"
			$cmd
		fi
	done
}


#------------------------------------------------------------------------------
#
# target
#
#------------------------------------------------------------------------------
function target
{
	FN="iSCSI_TARGET"
	
	local target
	local ip_allow
	
	local cmd
	local tid
	
	target="$1"
	ip_allow="$2"
	
	if [ -z "$target" -o -z "$ip_allow" ]; then
		echo "ERR> line: target $*"
		echo "ERR> calling target with incorrect number of args (target target_name ip_allow)"
		exit 1
	fi
	
	target_name=$TARGET_PREFIX:$target
	echo "$FN> target_name: $target_name"
	echo "$FN> ip_allow: $ip_allow"
	
	if [ $run_mode = "online" ]; then
		#
		# check if there is a tid for this target
		#
		tid=$(get_tid_from_target "$target_name")
		ccode=$?
		if [ $ccode = 1 ]; then
			echo "ERR> error invoking get_tid_from_target with target_name: $target_name"
			exit 1
		fi
		echo "$FN> tid: $tid"
	
		if [ "$tid" = 0 ]; then
			tid=$(get_next_tid)
			echo "$FN> since tid is 0, looking for next tid: $tid"
		
			cmd="ietadm --op new --tid=$tid --params Name=$target_name"
			echo "$FN> $cmd"
			$cmd
			ccode=$?
			if [ $ccode != 0 ]; then
				echo "ERR> ietadm returned error code $ccode"
				exit 1
			fi
		fi
	
		#
		# remember the tid for subsequent functions
		#
		g_tid="$tid"
		g_tid_ip_allow="$ip_allow"
	fi
	
	#
	# write out the target definition into the config file
	#
	echo "" >> $CONF_IETD
	echo "Target $target_name" >> $CONF_IETD
	
	#
	# invoke common options for target, they can be overridden by the
	# target's option statements
	#
	# Note: the alias was always broken (i.e. it never shows up
	# at the initiator
	#option Alias "$target"
	target_common_option
	
	#
	# write out the access control entries
	#
	echo "$target_name ALL" >> $CONF_DENY
	echo "$target_name $ip_allow" >> $CONF_ALLOW
}

#------------------------------------------------------------------------------
#
# lun (only fileio type luns are supported!)
#
#------------------------------------------------------------------------------
function lun
{
	FN="iSCSI_LUN"
	
	local lun_id
	local lun_path
	local lun_aux
	
	lun_id="$1"
	lun_path="$2"
	lun_aux="$3"

	if [ -z "$lun_id" -o -z "$lun_path" ]; then
		echo "ERR> calling lun with incorrect number of args (lun lun_id lun_path ietd_parms)"
		exit 1
	fi
	
	echo "$FN> conf lun: $lun_id, path: $lun_path, aux: $lun_aux"
	
	if [ -n "$lun_aux" ]; then
		lun_aux=",$lun_aux"
	fi
	
	if [ $run_mode = "online" ]; then
		#
		# check lun path (only in online mode - when offline, it is not clear if the
		# paths are valid)
		#
		if [ ! -e $lun_path ]; then
			echo "ERR> lun path: $lun_path not found."
			exit 1
		fi
		
		#
		# check if lun number is already assigned, if so, do nothing, else
		# create new lun. Lun deletions must be done from the command line!
		#
		luns_info=$(get_lun_info_from_tid "$g_tid")

		lun_info=$(echo "$luns_info" | grep "lun:$lun_id ")
		if [ -z "$lun_info" ]; then
			#
			# create new lun
			#
			echo "$FN> lun $lun_id not found, creating it"
			cmd="ietadm --op new --tid=$g_tid --lun=$lun_id --params=Path=$lun_path,Type=$LUN_TYPEIO"$lun_aux
			echo "$FN> $cmd"
			$cmd
			ccode=$?
			if [ $ccode != 0 ]; then
				echo "ERR> ietadm returned error code $ccode"
				exit 1
			fi
			
			#
			# track the size of the lun (LVM only supported)
			#
			lun_size=$(get_lv_size "$lun_path")
			track_lun_size_set "$lun_path" "$lun_size"
			echo "$FN> size: $lun_size"
		else
			#
			# check out the state of the current lun
			#
			
			#         lun:0 state:0 iotype:fileio iomode:wt sn:... path:/dev/vg_atabeast0_vol0/lv_as3test
			ietd_lun_path=$(echo "$lun_info" | awk '{print $NF'} | cut -f2 -d:)
			echo "$FN> ietd lun: $lun_id, path: $ietd_lun_path (UP)"
			
			#
			# verify that the path agrees between the config and active system
			#
			if [ "$lun_path" != "$ietd_lun_path" ]; then
				echo "ERR> lun path mismatch, you must remove the active LUN manually using ietadm"
				exit 1
			fi
			
			#
			# obtain the tracked volume size (if there is none, then track it now)
			# if there is one, compare it to the current size and if that is different
			# we need to do a ietd fixup to have it recognize the new volume size
			#
			lun_size=$(get_lv_size "$lun_path")
			tracked_lun_size=$(track_lun_size_get "$lun_path")
			if [ -z "$tracked_lun_size" ]; then
				echo "$FN> tracked lun size not found, lun is $lun_size, tracking it now (OK)"
				track_lun_size_set "$lun_path" "$lun_size"
			else
				#
				# check of the lun_size is different from the tracked_lun_size
				#
				if  [ "$lun_size" != "$tracked_lun_size" ]; then
					echo "$FN> lun size was $tracked_lun_size, current $lun_size (UPDATING)"

					echo "--BEGIN LUN RESIZE UPDATE"
					
					#
					# block tcp stream for this lun's target's ip allow addresses
					#
					target_io "$g_tid_ip_allow" "block"
					
					#
					# drop all active sessions from this target
					#
					delete_target_connections "$g_tid"
					
					#
					# remove lun from ietd
					#
					cmd="ietadm --op delete --tid=$g_tid --lun=$lun_id"
					echo "$FN> $cmd"
					$cmd
					ccode=$?
					if [ $ccode != 0 ]; then
						echo "ERR> ietadm returned error code $ccode"
						exit 1
					fi
					sleep 5
					
					#
					# add lun to ietd
					#
					cmd="ietadm --op new --tid=$g_tid --lun=$lun_id --params=Path=$lun_path,Type=$LUN_TYPEIO"$lun_aux
					echo "$FN> $cmd"
					$cmd
					ccode=$?
					if [ $ccode != 0 ]; then
						echo "ERR> ietadm returned error code $ccode"
						exit 1
					fi
					sleep 5

					#
					# unblock tcp stream (this will cause re-connections to target)
					#
					target_io "$g_tid_ip_allow" "unblock"
					
					#
					# track the size of the lun (LVM only supported)
					#
					lun_size=$(get_lv_size "$lun_path")
					track_lun_size_set "$lun_path" "$lun_size"
					
					echo "--END LUN RESIZE UPDATE"
				fi
				echo "$FN> size: $lun_size"
			fi
		fi
	fi

	echo "	Lun $lun_id Path=$lun_path,Type=$LUN_TYPEIO"$lun_aux >> $CONF_IETD
	
	lun_callback "$lun_path"
}

#------------------------------------------------------------------------------
#
# option
#
#------------------------------------------------------------------------------
function option
{
	FN="iSCSI_OPTION"
	
	local option_name
	local option_val
	
	option_name="$1"
	option_val="$2"
	
	local cmd
	
	if [ -z "$option_name" -o -z "$option_val" ]; then
		echo "ERR> calling option with incorrect number of args (option name value)"
		exit 1
	fi
	
	if [ "$g_option_state" = "global" ]; then
		echo "$FN> global option: $option_name=$option_val"
		echo "$option_name $option_val" >> $CONF_IETD
	else
		if [ $DEBUG_OPTION == 1 ]; then
			echo "$FN> local option: $option_name=$option_val"
		fi
		if [ $run_mode = "online" ]; then
			cmd="ietadm --op update --tid=$g_tid --params=$option_name=$option_val"
			if [ $DEBUG_OPTION == 1 ]; then
				echo "$FN> $cmd"
			fi
			$cmd
			ccode=$?
			if [ $ccode != 0 ]; then
				echo "ERR> ietadm returned error code $ccode"
				exit 1
			fi
		fi
		echo "	$option_name $option_val" >> $CONF_IETD
	fi
}

cat << GLOBAL > $CONF_IETD
#-----------------------------------------------
#
# iSCSI Enterprise Target Configuration
#
# Caution: do not edit, edit master file instead
# located at `pwd`/target.conf
#
# Generated by: `uname -n`
# Generated on: `date`
#
# Changes are dynamically added (add only) to
# ietd using ietadm. You are required to remove
# targets, luns, session using ietadm.
#
#-----------------------------------------------
GLOBAL

cat << ALLOW > $CONF_ALLOW
#-----------------------------------------------
#
# Examples of initiators that are allowed to connect
#
# iqn.2001-04.com.example:storage.disk1.sys1.xyz 192.168.22.2, 192.168.3.8
# iqn.2001-04.com.example:storage.disk1.sys4.xyz [3ffe:302:11:1:211:43ff:fe31:5ae2], [3ffe:505:2:1::]/64
#-----------------------------------------------
ALLOW



cat << DENY > $CONF_DENY
#-----------------------------------------------
#
# Examples of initiators that are denied
#
# iqn.2001-04.com.example:storage.disk1.sys1.xyz ALL
# iqn.2001-04.com.example:storage.disk1.sys2.xyz 192.168.12.2, 192.168.3.0/24, 192.167.1.16/28
# iqn.2001-04.com.example:storage.disk1.sys4.xyz [3ffe:302:11:1:211:43ff:fe31:5ae2], [3ffe:505:2:1::]/64
#-----------------------------------------------
DENY

#
# invoke all global options (global option changes, 
# require iscsi-target restart)
#
g_option_state="global"
global_options

#
# change future options to target (localized) options
# local options are pushed via ietadm into the current environment
#
g_option_state="local"
g_tid=0
g_tid_ip_allow=""

lun_callback_pre
build_targets_luns
lun_callback_post

#cat $CONF_IETD #$CONF_ALLOW $CONF_DENY
